from __future__ import annotations

import json
import time
from math import modf
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Iterable

import download_youtube_subtitle.main as down_yt_subs

__all__ = ('download_srt',)


def find_caption_num(video_id: str, lang_code: str) -> int | None:
    video_id = down_yt_subs.parseVideoID(video_id)[0]
    try:
        caption_tracks = down_yt_subs.get_tracks_title(video_id)[0]
    except Exception:
        down_yt_subs.perr("can't retrive caption")
        raise
    for i, caption in enumerate(caption_tracks):
        if caption.get('languageCode') == lang_code:
            return i


subs_download = down_yt_subs.main


def _float_to_srt_time_format(d: float) -> str:
    fraction, whole = modf(d)
    return f"{time.strftime('%H:%M:%S,', time.gmtime(whole))}{f'{fraction:.3f}'.replace('0.', '')}"


def _from_json(translation: Iterable[dict[str, str]]) -> str:
    return '\n\n'.join(
        f"{i}\n{_float_to_srt_time_format((start := int(el['start'])) / 1000.)}"
        " --> "
        f"{_float_to_srt_time_format((start + int(el['dur'])) / 1000.)}\n{el['text']}"
        for i, el in enumerate(translation, 1))


def download_srt(video_id: str, lang: str, fp: str | Path):
    with TemporaryDirectory() as tmp:
        tmp_name = str(Path(tmp) / f'{video_id}_{lang}.srt.json')
        cap_num = find_caption_num(video_id, lang)
        if cap_num is None:
            subs_download(video_id, lang, output_file=tmp_name, to_json=True)
        else:
            subs_download(video_id, False, caption_num_second=cap_num,
                          output_file=tmp_name, to_json=True)
        with open(tmp_name, encoding='UTF-8') as f:
            _json: dict[str, list[dict[str, str]]] = json.load(f)
    with open(fp, 'w', encoding='UTF-8') as f:
        f.write(_from_json(_json.get('translation', ())))



__all__ = 'SUPPORTED',
try:
    import translators
    from fastdub.translator import subs_translate

    SERVICES = [i[1:] for i in dir(translators.apis) if i.startswith('_') and not i.startswith('__')]


    def get_service_by_name(service: str):
        return getattr(translators, service)
except ImportError:
    SUPPORTED = False
else:
    SUPPORTED = True
    __all__ += ('translators', 'SERVICES', 'get_service_by_name', 'subs_translate')



"""Convert JSON file with auto-generated subtitles from https://www.veed.io to .srt file"""
import argparse
import json
from pathlib import Path

from fastdub.subtitles import ms_to_srt_time

__all__ = ('from_veed_json', 'main')


def from_veed_json(json_file: Path):
    with json_file.open('r', encoding='UTF-8') as f:
        jls = json.load(f)
    (json_file.parent / f'{json_file.stem}.srt').write_text('\n\n'.join(
        f"{i}\n{ms_to_srt_time(v['from'] * 1000)} --> {ms_to_srt_time(v['to'] * 1000)}\n"
        f"{' '.join(t['value'] for t in v['words'])}"
        for i, v in enumerate((*(jls['data']['edit']['subtitles']['tracks'].values()),)[0]['items'].values(), 1)
    ), 'UTF-8')


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('json_file', type=Path)
    args = parser.parse_args()
    from_veed_json(args.json_file)


if __name__ == '__main__':
    main()



try:
    from fastdub.youtube.yt_upload import uploader
    from fastdub.youtube.yt_upload.upload_video import VALID_PRIVACY_STATUSES
except ImportError:
    SUPPORTED = False
else:
    SUPPORTED = True



from __future__ import annotations

import os.path
from contextlib import closing
from pathlib import Path
from typing import Sequence

import rich.align
from moviepy.video.io.VideoFileClip import VideoFileClip
from tqdm import tqdm

from fastdub import audio, subtitles, voicer
from fastdub.ffmpeg_wrapper import FFMpegWrapper

__all__ = ('Dubber', 'VOICER')

VOICER = voicer.Voicer()


class Dubber:
    __slots__ = (
        'fit_align', 'language', 'audio_format',
        'ducking', 'min_silence_len', 'silence_thresh',
        'gain_during_overlay',
        'cleanup_level'
    )

    def __init__(self, voice: str, language: str, audio_format: str,
                 ducking: bool, min_silence_len: int, silence_thresh: float, gain_during_overlay: int,
                 fit_align: float = 2., cleanup_level: int = 2):
        self.language = language
        self.audio_format = audio_format
        self.fit_align = fit_align
        if voice:
            VOICER.set_voice(voice)

        self.ducking = ducking
        self.min_silence_len = min_silence_len
        self.silence_thresh = silence_thresh
        self.gain_during_overlay = gain_during_overlay

        self.cleanup_level = cleanup_level

    @staticmethod
    def collect_videos(path_to_files: str, skip_starts_underscore: bool = True,
                       exclude_files: Sequence[str] = frozenset()):
        path_to_files = os.path.abspath(path_to_files)
        videos = {}
        for file in os.listdir(path_to_files):
            if file in exclude_files:
                continue
            absfile = os.path.join(path_to_files, file)
            if not os.path.isfile(absfile) or skip_starts_underscore and file.startswith('_'):
                continue
            filename, ext = os.path.splitext(file)
            if videos.get(filename):
                videos[filename][ext] = absfile
                continue
            videos[filename] = {ext: os.path.join(path_to_files, file)}
        return videos

    def dub_dir(self, videos: dict[str, dict[str, str]], video_format: str, subtitles_format: str):
        for fn, exts in videos.items():
            target_vid = exts.get(video_format)
            target_sub = exts.get(subtitles_format)

            if not target_vid:
                raise FileNotFoundError(fn + video_format)
            if not target_sub:
                raise FileNotFoundError(fn + subtitles_format)

            self.dub_one(fn, target_vid, target_sub)

    def dub_one(self, fn: str, target_vid: str, target_sub: str, cleanup_audio: bool = None):
        if target_vid is target_sub is None:
            return
        rich.print(rich.align.Align(fn, 'center'))
        subs = subtitles.parse(target_sub)

        default_right_border = 0
        if target_vid and subs:
            video_clip: VideoFileClip
            with closing(VideoFileClip(target_vid, audio=False)) as video_clip:
                default_right_border = video_clip.end * 1000. - subs[-1].ms.end
        default_right_border = max(default_right_border, 0)

        progress_total = len(subs)

        fit_align = self.fit_align
        cur_audio = audio.AudioSegment.empty()
        tts_list = *(VOICER.voice(line.text) for line in
                     tqdm(subs, desc='TTSing', total=progress_total,
                          unit='line', dynamic_ncols=True)),
        for pos, tts in tqdm(enumerate(tts_list[:-1]),
                             desc='Fitting audio',
                             total=progress_total - 1, unit='line',
                             dynamic_ncols=True):
            ms = subs[pos].ms
            cur_audio += audio.fit(
                tts,
                ms.start - cur_audio.duration_ms,
                ms.duration,
                subs[pos + 1].ms.start - ms.end,
                fit_align
            )
        ms = subs[- 1].ms
        cur_audio += audio.fit(tts_list[-1],
                               ms.start - cur_audio.duration_ms,
                               ms.duration,
                               default_right_border,
                               fit_align)

        max_duration = ms.end
        audio_duration = cur_audio.duration_ms

        if audio_duration > max_duration:
            change_speed = audio_duration / max_duration
            rich.print(f'Changing audio speed to {change_speed:g}')
            cur_audio = audio.speed_change(cur_audio, change_speed)
        elif audio_duration != max_duration:
            cur_audio = audio.AudioSegment.silent(
                min(max_duration - audio_duration, subs[0].ms.start)
            ) + cur_audio

        result_dir = Path(target_sub).parent / '_result'
        result_dir.mkdir(exist_ok=True)
        result_out_audio = str(result_dir / f'{fn}_{self.language}.{self.audio_format}')

        if cleanup_audio is None:
            cleanup_audio = self.cleanup_level

        if target_vid:
            if self.ducking:
                audio.side_chain(audio.AudioSegment.from_file(target_vid),
                                 cur_audio,
                                 self.min_silence_len, self.silence_thresh, self.gain_during_overlay
                                 ).export(result_out_audio)
            else:
                audio.AudioSegment.from_file(target_vid).overlay(cur_audio,
                                                                 gain_during_overlay=self.gain_during_overlay
                                                                 ).export(result_out_audio)

            FFMpegWrapper.save_result_data(target_vid, result_out_audio, target_sub,
                                           result_dir / f'{fn}_{self.language}.mp4')
            if cleanup_audio:
                os.remove(result_out_audio)
        else:
            cur_audio.export(result_out_audio)



# https://developers.google.com/youtube/v3/guides/uploading_a_video/
import argparse
import http.client
import json
import logging
import random
import sys
import time
import webbrowser
from pathlib import Path

import httplib2
# noinspection PyUnresolvedReferences
from apiclient.discovery import build
# noinspection PyUnresolvedReferences
from apiclient.errors import HttpError
# noinspection PyUnresolvedReferences
from apiclient.http import MediaFileUpload
from oauth2client.client import flow_from_clientsecrets
from oauth2client.file import Storage
from oauth2client.tools import run_flow

httplib2.RETRIES = 1

MAX_RETRIES = 10

RETRIABLE_EXCEPTIONS = (httplib2.HttpLib2Error, IOError, http.client.NotConnected,
                        http.client.IncompleteRead, http.client.ImproperConnectionState,
                        http.client.CannotSendRequest, http.client.CannotSendHeader,
                        http.client.ResponseNotReady, http.client.BadStatusLine)

RETRIABLE_STATUS_CODES = {500, 502, 503, 504}

CLIENT_SECRETS_FILE = 'client_secrets.json'

YOUTUBE_UPLOAD_SCOPE = 'https://www.googleapis.com/auth/youtube.upload'
YOUTUBE_API_SERVICE_NAME = 'youtube'
YOUTUBE_API_VERSION = 'v3'

MISSING_CLIENT_SECRETS_MESSAGE = f'''\
WARNING: Please configure OAuth 2.0

To make this sample run you will need to populate the client_secrets.json file
found at:

   {(Path(__file__).parent / CLIENT_SECRETS_FILE).absolute()}

with information from the API Console
https://console.developers.google.com/

For more information about the client_secrets.json file format, please visit:
https://developers.google.com/api-client-library/python/guide/aaa_client_secrets'''

VALID_PRIVACY_STATUSES = ('private', 'public', 'unlisted')


def get_authenticated_service(args):
    flow = flow_from_clientsecrets(CLIENT_SECRETS_FILE,
                                   scope=YOUTUBE_UPLOAD_SCOPE,
                                   message=MISSING_CLIENT_SECRETS_MESSAGE)

    storage = Storage(f'{sys.argv[0]}-oauth2.json')
    credentials = storage.get()

    if credentials is None or credentials.invalid:
        credentials = run_flow(flow, storage, args)

    return build(YOUTUBE_API_SERVICE_NAME, YOUTUBE_API_VERSION,
                 http=credentials.authorize(httplib2.Http()))


def initialize_upload(youtube, options):
    tags = None
    if options.keywords:
        tags = options.keywords.split(',')

    body = {'snippet': {'title': options.title,
                        'description': options.description,
                        'tags': tags,
                        'categoryId': options.category},
            'status': {'privacyStatus': options.privacyStatus}}

    # Call the API's videos.insert method to create and upload the video.
    insert_request = youtube.videos().insert(
        part=','.join(body.keys()),
        body=body,
        media_body=MediaFileUpload(options.file, chunksize=-1, resumable=True)
    )

    resumable_upload(insert_request)


def _json_content(content):
    return json.dumps(json.loads(content.decode('UTF-8')), ensure_ascii=False, indent=1)


def resumable_upload(insert_request):
    response = None
    error = None
    retry = 0
    while response is None:
        try:
            logging.info('Uploading file...')
            status, response = insert_request.next_chunk()
            if response is not None:
                if 'id' in response:
                    id_ = response['id']
                    try:
                        webbrowser.open_new_tab(f'https://youtu.be/{id_}')
                    except webbrowser.Error:
                        pass
                    logging.info(f'Video id {id_!r} was successfully uploaded.')
                else:
                    exit(f'The upload failed with an unexpected response: {response}')
        except HttpError as e:
            if e.resp.status in RETRIABLE_STATUS_CODES:
                error = f'A retriable HTTP error {e.resp.status} occurred:\n{_json_content(e.content)}'
            else:
                raise
        except RETRIABLE_EXCEPTIONS as e:
            error = f'A retriable error occurred: {e}'

        if error is not None:
            logging.error(error)
            retry += 1
            if retry > MAX_RETRIES:
                exit('No longer attempting to retry.')

            max_sleep = 2**retry
            sleep_seconds = random.random() * max_sleep
            logging.info(f'Sleeping {sleep_seconds:g} seconds and then retrying...')
            time.sleep(sleep_seconds)


def parse_args(args=None):
    argparser = argparse.ArgumentParser(add_help=False)
    argparser.add_argument('--auth_host_name', default='localhost',
                           help='Hostname when running a local web server.')
    argparser.add_argument('--noauth_local_webserver', action='store_true',
                           default=False, help='Do not run a local web server.')
    argparser.add_argument('--auth_host_port', default=(8080, 8090), type=int,
                           nargs='*', help='Port web server should listen on.')
    argparser.add_argument(
        '--logging_level', default='ERROR',
        choices=('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'),
        help='Set the logging level of detail.')

    argparser.add_argument('--file', type=Path, required=True, help='Video file to upload')
    argparser.add_argument('--title', help='Video title', default='Untitled')
    argparser.add_argument('--description', help='Video description',
                           default='')
    argparser.add_argument('--category', default='22',
                           help='Numeric video category. ' +
                                'See https://developers.google.com/youtube/v3/docs/videoCategories/list')
    argparser.add_argument('--keywords', help='Video keywords, comma separated',
                           default='')
    argparser.add_argument('--privacy-status', choices=VALID_PRIVACY_STATUSES,
                           default=VALID_PRIVACY_STATUSES[0], help='Video privacy status.', dest='privacyStatus')
    return argparser.parse_args(args)


def upload(args=None):
    args = parse_args(args)
    if not args.file.is_file():
        exit('Please specify a valid file using the --file parameter.')

    youtube = get_authenticated_service(args)
    try:
        initialize_upload(youtube, args)
    except HttpError as e:
        logging.error(f'An HTTP error {e.resp.status} occurred:\n{_json_content(e.content)}')



from __future__ import annotations

import multiprocessing.pool
import re
from pathlib import Path
from typing import Callable, Optional, TypeVar
from urllib.parse import urlparse

import rich.live
import rich.table
from tqdm import tqdm
from youtubesearchpython import VideosSearch

from fastdub import PrettyViewPrefix
from fastdub.youtube import *
from fastdub.youtube import pafy
from fastdub.youtube.pafy import g as pafy_g
from fastdub.youtube.subtitles import download_srt

__all__ = ('DownloadYTVideo', 'with_api_key')

_API_RET_TYPE = TypeVar('_API_RET_TYPE')
_PATH_SUPPORTED = re.compile(r'[\\/:?"<>|]+')


def _path_save(name: str) -> str:
    return _PATH_SUPPORTED.sub('_', name)


class DownloadYTVideo:
    __slots__ = ('save_dir', 'language', 'playlist', 'api_keys',
                 'download_title',
                 '_table_data', '_live_progress')
    API_KEYS = {pafy_g.api_key, 'AIzaSyCHxJ84-ryessLJfWZVWldiuVCnxtf0Nm4'}
    TABLE_TITLE_TEMPLATE = 'Downloading from [link=%(scheme)s://%(netloc)s]%(netloc)s[/]...'

    def __init__(self, query: str,
                 language: str,
                 search_limit: int = 20,
                 region: str = 'US'):
        url_parsed = urlparse(query)
        url_path = url_parsed.path
        is_youtube = False
        if query.startswith('?'):
            query = query.removeprefix('?')
            is_youtube = True
            videos = tqdm(VideosSearch(query, search_limit, language, region).result().get('result', ()),
                          'Video search processing', unit='video', dynamic_ncols=True, colour='white')
            playlist = (
                *(with_api_key(lambda: pafy.new(data['id']))
                  for data in videos if data.get('type', '') == 'video'),)
            save_dir = query
        elif url_path == '/playlist':
            playlist = with_api_key(lambda: pafy.get_playlist2(query))
            save_dir = f'{playlist.title} [{playlist.plid}]'
        elif (path_split := url_path.strip('/').split('/')) and path_split[0] in ('c', 'channel'):
            playlist = with_api_key(lambda: pafy.get_channel((path_split[1:] or (query,))[0]).uploads)
            save_dir = f'{playlist.title} [{playlist.plid}]'
        else:
            playlist = with_api_key(lambda: (pafy.new(query),))
            playlist_0 = playlist[0]
            save_dir = f'{playlist_0.title} [{playlist_0.videoid}]'
        save_dir = Path(_path_save(' '.join(save_dir.split(' ')[:-1]) if len(save_dir) > 100 else save_dir))
        save_dir.mkdir(parents=True, exist_ok=True)

        self.save_dir = save_dir
        self.language = language
        self.playlist = playlist
        self.download_title = self.TABLE_TITLE_TEMPLATE % (
            {'scheme': 'https', 'netloc': 'www.youtube.com'} if is_youtube
            else {'scheme': url_parsed.scheme, 'netloc': url_parsed.netloc}
        )

    def download(self, yt_dl: YtdlPafy):
        title: str = yt_dl.title.strip()
        if len(title) <= (97 - len(yt_dl.videoid)):
            title += f' [{yt_dl.videoid}]'
        save_to = self.save_dir / title

        _file = Path(save_to)
        srt_file = _file.with_suffix('.srt')
        if not srt_file.is_file():
            download_srt(yt_dl.videoid, self.language, srt_file)
        mp4_file = _file.with_suffix('.mp4')
        if not mp4_file.is_file():
            try:
                with_api_key(self.mp4_downloader(yt_dl, mp4_file))
            except OSError:
                pass

    def mp4_downloader(self, yt_dl: YtdlPafy, mp4_file: str | Path):
        mp4_file = str(mp4_file)
        return lambda: yt_dl.getbest('mp4').download(mp4_file,
                                                     quiet=True,
                                                     callback=lambda *args: self.progress_callback(mp4_file, *args))

    def multiprocessing_download(self, pc: int = None):
        self._table_data = {}
        self._live_progress = rich.live.Live(self._generate_info_table())
        with self._live_progress:
            if pc is None:
                pc = multiprocessing.cpu_count()
            if pc < 2:
                for yt_dl in self.playlist:
                    self.download(yt_dl)
                return
            with multiprocessing.pool.ThreadPool(pc) as pool:
                pool.map(self.download, self.playlist)
            self._live_progress.update(f'{len(self.playlist)} downloaded.')

    def progress_callback(self, fn: str, total: int, downloaded: float, ratio: float, rate: float, eta: float):
        if ratio == 1.:
            if fn in self._table_data:
                del self._table_data[fn]
            return
        self._table_data[fn] = (
            f'\r{ratio:.2%}',
            f'{PrettyViewPrefix.from_bytes(downloaded)}/{PrettyViewPrefix.from_bytes(total)}',
            f'{PrettyViewPrefix.from_bytes(rate)}/s', PrettyViewPrefix.from_seconds(eta)
        )
        self._live_progress.update(self._generate_info_table())

    def _generate_info_table(self):
        _download_table = rich.table.Table(title=self.download_title,
                                           show_lines=True,
                                           highlight=True)
        for header in ('file name', '%', 'downloaded/total', 'rate', 'ETA'):
            _download_table.add_column(header, justify='center')
        for fn, v in self._table_data.items():
            _download_table.add_row(fn, *v)
        return _download_table


def with_api_key(func: Callable[[], _API_RET_TYPE]) -> Optional[_API_RET_TYPE]:
    for key in DownloadYTVideo.API_KEYS:
        pafy.set_api_key(key)
        try:
            return func()
        except pafy.util.GdataError:
            continue



from __future__ import annotations

from subprocess import check_call

from imageio_ffmpeg import get_ffmpeg_exe

__all__ = ('FFMpegWrapper',)


class FFMpegWrapper:
    __slots__ = ()
    DEFAULT_FFMPEG_LOG_LEVEL = 'panic'
    DEFAULT_ARGS = ()
    FFMPEG_EXE = get_ffmpeg_exe()

    @classmethod
    def convert(cls, *args, loglevel=None):
        if not loglevel:
            loglevel = cls.DEFAULT_FFMPEG_LOG_LEVEL
        check_call((cls.FFMPEG_EXE, '-v', loglevel, *cls.DEFAULT_ARGS, *args))

    @staticmethod
    def save_result_data(video_path, audio_path, subtitles_path, output_path):
        FFMpegWrapper.convert('-i', video_path, '-i', audio_path, '-i', subtitles_path,
                              '-map', '0:0', '-map', '1:0', '-map', '0:1', '-map', '2:0',
                              '-disposition:a:0', 'default', '-disposition:a:1', '0',
                              '-c', 'copy', '-c:s', 'mov_text',
                              output_path)



from __future__ import annotations

import argparse
from glob import iglob
from math import inf
from multiprocessing import cpu_count
from os import getcwd
from time import perf_counter

import rich.traceback

import fastdub.youtube
from fastdub import PrettyViewPrefix, dubber, translator, voicer, youtube
from fastdub.ffmpeg_wrapper import FFMpegWrapper
from fastdub.translator.subs_translate import SrtTranslate

__all__ = ('parse_args', 'main')


class BooleanOptionalAction(argparse.Action):
    __slots__ = ()

    def __init__(self, option_strings, dest, default=None, type=None,
                 choices=None, required=False, help=None, metavar=None):
        _option_strings = (*option_strings,
                           *('--no-' + option_string[2:] if option_string.startswith('--') else
                             '-n-' + option_string[1:] if option_string.startswith('-') else option_strings
                             for option_string in option_strings))
        if help is not None and default is not None and default is not argparse.SUPPRESS:
            help += " (default: %(default)s)"
        super().__init__(option_strings=_option_strings, dest=dest, nargs=0, default=default, type=type,
                         choices=choices, required=required, help=help, metavar=metavar)

    def __call__(self, parser, namespace, values, option_string=None):
        if option_string in self.option_strings:
            setattr(namespace, self.dest,
                    not (option_string.startswith('--no-') or option_string.startswith('-n-')))

    def format_usage(self):
        return ' | '.join(self.option_strings)


# noinspection PyTypeChecker
def parse_args() -> argparse.Namespace:
    arg_parser = argparse.ArgumentParser('fastdub',
                                         description='fastdub is a tool for dubbing videos by subtitle files.',
                                         formatter_class=argparse.RawTextHelpFormatter)

    arg_parser.add_argument('-rc', '--remove-cache', default=2, type=int, choices=(0, 1, 2),
                            help='Remove all cache (_cached_texts directory) files\n'
                                 '\t0 - No remove cache\n'
                                 '\t1 - Delete cache before voice acting\n'
                                 '\t2 - Delete cache after voice acting (default)')
    arg_parser.add_argument('-ra', '--cleanup-audio', action=BooleanOptionalAction, default=False,
                            help='Remove result audio if video exists', dest='cleanup_audio')

    arg_parser.add_argument('-l', '--language', default='ru',
                            help='Subtitles language (ru)')

    if youtube.SUPPORTED or translator.SUPPORTED:
        arg_parser.add_argument('-tc', '--threads-count', default=cpu_count(),
                                type=lambda i: int(cpu_count() * float(i.lstrip('*')))
                                if i.startswith('*') else int(i),
                                help='Process count to download (pass to cpu count, < 2 to disable)\n'
                                     '\t*N = N * cpu count')

    input_group = arg_parser.add_argument_group('Input')
    input_group.add_argument('-i', '--input', default=getcwd(), required=True,
                             help=f"Input directory{('/YouTube Playlist/Video URL' if youtube.SUPPORTED else '')}.")
    input_group.add_argument('-vf', '--video-format', default='.mp4',
                             help='Video format (default: .mp4).')
    input_group.add_argument('-sf', '--subtitles-format', default='.srt',
                             help='Subtitles format (default: .srt).')

    exclude_input_group = input_group.add_argument_group('Exclude Input')
    exclude_input_group.add_argument('-En', '--exclude', type=iglob, nargs='+', default=(),
                                     help='Exclude <name> (glob)')
    exclude_input_group.add_argument('-Eu', '--exclude-underscore', default=True,
                                     help='Exclude files starts with underscore')

    ducking_group = arg_parser.add_argument_group('Audio Ducking')
    ducking_group.add_argument('-sc', '--side-chain', action=BooleanOptionalAction, default=True,
                               help='Enable audio side chain compress (ducking)')
    ducking_group.add_argument('-sc-msl', '--min-silence-len', '--attack', default=100, type=int,
                               help='Minimum silence length in ms (default: 100)')
    ducking_group.add_argument('-sc-st', '--silence-thresh', default=-inf, type=float,
                               help='Silence threshold in dB')
    ducking_group.add_argument('-sc-gdo', '--gain-during-overlay', default=-11, type=int,
                               help='Gain during overlay in dB (default: -11)')

    voicer_group = arg_parser.add_argument_group('Voicer')
    voicer_group.add_argument('-v', '--voice', type=str.lower, choices=voicer.VOICES_NAMES.keys(),
                              help='SAPI voice for voice acting.')
    voicer_group.add_argument('-a', '--align', default=2., type=float,
                              help='Audio fit align (divisor)'
                                   '\n\t1 = right'
                                   '\n\t2 = center (default)')
    voicer_group.add_argument('-v-set-a', '--voice-set-anchor', default='!:',
                              help='Anchor indicating voice actor change (default "!:")')

    ffmpeg_group = arg_parser.add_argument_group('FFMpeg Output')
    ffmpeg_group.add_argument('-ll', '--loglevel', default='panic',
                              choices=(
                                  'trace', 'debug', 'verbose', 'info', 'warning', 'error', 'fatal', 'panic', 'quiet'),
                              help='FFMpegWrapper loglevel')
    ffmpeg_group.add_argument('-y', '--confirm', action=BooleanOptionalAction, default=True,
                              help="Don't ask for confirmation")
    ffmpeg_group.add_argument('-af', '--audio-format', default='wav',
                              help='Out audio files format (default wav)')

    output_group = arg_parser.add_argument_group('Terminal Output')
    output_group.add_argument('-tb', '--traceback', action=BooleanOptionalAction, default=False,
                              help='Show debug traceback')

    if youtube.SUPPORTED:
        yt_group = arg_parser.add_argument_group('YouTube')
        yt_group.add_argument('-yt', '--youtube', action='store_true', default=False)
        yt_group.add_argument('-ak', '--api-keys', nargs='+', default=(), help='Youtube API key/s')

        yt_search_group = arg_parser.add_argument_group('YouTube Search')
        yt_search_group.add_argument('-yts', '--youtube-search', action='store_true', default=False,
                                     help='The input data is a query for searching on YouTube.'
                                          ' (Adds "?" at the start of input)')
        yt_search_group.add_argument('-yts-l', '--youtube-search-limit', type=int, default=20,
                                     help='Sets limit to the number of results. Defaults to 20.')
        yt_search_group.add_argument('-yts-rg', '--youtube-search-region', default='US',
                                     help='Sets the result region. Defaults to "US".')

    if fastdub.youtube.yt_upload.SUPPORTED:
        ytu_group = arg_parser.add_argument_group('YouTube Upload')
        ytu_group.add_argument('-ytu', '--youtube-upload', action='store_true', default=False,
                               help='yt_upload video to YouTube channel after voice acting.')
        ytu_group.add_argument('-ytu-ps', '--privacy-status',
                               default=fastdub.youtube.yt_upload.VALID_PRIVACY_STATUSES[0],
                               choices=fastdub.youtube.yt_upload.VALID_PRIVACY_STATUSES,
                               help='Video privacy status (If not private, errors are possible)')
        if translator.SUPPORTED:
            ytu_group.add_argument('-ytu-t', '--youtube-upload-translate', action='store_true', default=False,
                                   help='Translate title and description on upload. '
                                        '(+ Arguments from translate argument group)')

    if translator.SUPPORTED:
        translate_group = arg_parser.add_argument_group('Translate')
        translate_group.add_argument('-tr', '--translate', action='store_true', default=False,
                                     help='Translate input subtitles files')
        translate_group.add_argument('--rewrite-srt', action=BooleanOptionalAction, default=False,
                                     help='Rewrite input subtitles files.\n'
                                          'If not, add "_" to the beginning of the original subtitle file.')
        translate_group.add_argument('-ts', '--translate-service', type=translator.get_service_by_name,
                                     default='google',
                                     choices=translator.SERVICES,
                                     help='Subtitle translation service. (default google)')

    return arg_parser.parse_args()


def main():
    args = parse_args()
    if args.traceback:
        rich.traceback.install(show_locals=True)
    else:
        rich.traceback.install(extra_lines=0, word_wrap=True)
    remove_cache = args.remove_cache
    if remove_cache == 1:
        dubber.VOICER.cleanup()

    FFMpegWrapper.DEFAULT_FFMPEG_LOG_LEVEL = args.loglevel

    total_time = 0
    if args.confirm:
        FFMpegWrapper.DEFAULT_ARGS += '-y',
        total_time = perf_counter()

    if youtube.SUPPORTED and args.youtube:
        query: str = args.input
        if args.youtube_search and not query.startswith('?'):
            query = f'?{query.strip()}'
        youtube.downloader.DownloadYTVideo.API_KEYS |= {*args.api_keys}
        downloader = youtube.downloader.DownloadYTVideo(query,
                                                        args.language,
                                                        args.youtube_search_limit, args.youtube_search_limit, )
        downloader.multiprocessing_download(args.threads_count)
        args.input = downloader.save_dir

    video_format = video_format if (video_format := args.video_format).startswith('.') else f'.{video_format}'
    subtitles_format = (
        subtitles_format if (subtitles_format := args.subtitles_format).startswith('.') else f'.{subtitles_format}')
    audio_format: str = args.audio_format.removeprefix('.')

    videos = dubber.Dubber.collect_videos(args.input, args.exclude_underscore, (*sum(args.exclude, []),))

    if translator.SUPPORTED and args.translate:
        SrtTranslate(args.language, args.translate_service, args.rewrite_srt, args.threads_count
                     ).translate_dir(videos, subtitles_format)

    dubs = dubber.Dubber(args.voice, args.language, audio_format,
                         args.side_chain, args.min_silence_len, args.silence_thresh, args.gain_during_overlay,
                         args.align, args.cleanup_audio)

    dubs.dub_dir(videos, video_format, subtitles_format)

    if remove_cache == 2:
        voicer.Voicer().cleanup()

    if fastdub.youtube.yt_upload.SUPPORTED and args.youtube_upload:
        translate = False
        translate_serv = None
        if translator.SUPPORTED:
            translate = args.youtube_upload_translate
            translate_serv = args.translate_service
        fastdub.youtube.yt_upload.uploader.Uploader(args.privacy_status, translate, translate_serv).upload(args.input)

    if total_time:
        rich.print(f'Total time: {PrettyViewPrefix.from_seconds(perf_counter() - total_time)}')


if __name__ == '__main__':
    main()



from __future__ import annotations

from hashlib import md5
from os.path import isfile
from pathlib import Path
from shutil import rmtree

import pyttsx3

from fastdub.audio import AudioSegment

__all__ = ('UnknownVoice', 'VOICES_NAMES', 'VOICES_ID', 'Voicer')


class UnknownVoice(Exception):
    __slots__ = ()


_voices = pyttsx3.init().proxy.getProperty('voices')
VOICES_NAMES = {i.name.casefold(): i for i in _voices}
VOICES_ID = {i.id: i for i in _voices}
del _voices


class Voicer:
    __slots__ = ('engine', 'cache_dir', '_update_voice_anchor')

    def __init__(self, cache_dir: str = None, anchor: str = '!:'):
        if anchor:
            def _update_voice_anchor(line: str) -> bool:
                if line.startswith(anchor):
                    self.set_voice(line.removeprefix(anchor))
                    return True
                return False
        else:
            def _update_voice_anchor(_: str) -> bool:
                return False
        self._update_voice_anchor = _update_voice_anchor
        self.cache_dir = Path(__file__).parent / '_cached_texts' if cache_dir is None else Path(cache_dir)
        self.engine = pyttsx3.init()
        if not self.cache_dir.is_dir():
            self.cache_dir.mkdir(parents=True)

    def cleanup(self):
        rmtree(self.cache_dir)

    def set_voice(self, voice: str):
        voice_name = voice.casefold()
        voice_property = self.engine.proxy.getProperty('voice')
        voice = VOICES_NAMES.get(voice_name)
        if not voice:
            raise UnknownVoice(f'{voice_name} not in {str(tuple(VOICES_NAMES.keys()))}')
        if VOICES_ID.get(voice_property).name.casefold() != voice_name:
            self.engine.proxy.setProperty('voice', voice.id)

    def voice(self, text: str) -> AudioSegment:
        text = text.strip()
        if not text:
            return AudioSegment.silent(0)

        if self._update_voice_anchor((lines := text.splitlines())[0]):
            text = '\n'.join(lines[1:])

        cached = f'''{self.cache_dir / md5(f"{text}{self.engine.proxy.getProperty('voice')}".encode())
        .hexdigest()}.wav'''
        if not isfile(cached):
            self.engine.proxy.save_to_file(text, cached, None)
            self.engine.runAndWait()
        return AudioSegment.from_file(cached, format='wav')



from __future__ import annotations

import multiprocessing.pool
import os.path
from typing import Callable

from tqdm import tqdm

from fastdub import subtitles

try:
    from functools import cache
except ImportError:
    from functools import lru_cache


    def cache(func, /):
        return lru_cache(maxsize=None)(func)

__all__ = ('SrtTranslate',)


class SrtTranslate:
    __slots__ = ('language', 'rewrite', 'threads_count', 'service', '_pb')

    def __init__(self, language: str, service: Callable, rewrite: bool = False, threads_count: int = None):
        self.language = language
        self.service = service
        self.rewrite = rewrite
        self.threads_count = threads_count or multiprocessing.cpu_count()

    @cache
    def translate_line(self, text: str) -> str:
        return self.service(text, to_language=self.language)

    def translate_all(self, input_srt: str, output_srt: str):
        parsed = subtitles.parse(input_srt)
        for line in parsed:
            line.text = line.text.strip()
        _pb = tqdm(parsed, 'Translating', len(parsed), unit='line', dynamic_ncols=True)
        if self.threads_count > 1:
            def handler(line: subtitles.Line):
                line.text = self.translate_line(line.text)
                _pb.update()

            with multiprocessing.pool.ThreadPool(self.threads_count) as pool:
                pool.map(handler, parsed)
        else:
            for line in _pb:
                line.text = self.translate_line(line.text)
        with open(output_srt, 'w', encoding='UTF-8') as f:
            f.write(subtitles.unparse(parsed))

    def translate_dir(self, files: dict[str, dict[str, str]], subtitles_format: str):
        for fn, exts in files.items():
            inp = out = exts[subtitles_format]
            if not self.rewrite:
                inp = os.path.join(*os.path.split(inp)[:-1], f'_{os.path.basename(inp)}')
                if os.path.isfile(inp):
                    os.remove(inp)
                os.rename(out, inp)
            self.translate_all(inp, out)



from __future__ import annotations

import os.path
from array import array
from copy import copy
from math import inf
from tempfile import TemporaryDirectory

import pydub.silence
from tqdm import tqdm

from fastdub.ffmpeg_wrapper import FFMpegWrapper

__all__ = ('AudioSegment', 'speed_change', 'fit', 'side_chain')


class AudioSegment(pydub.AudioSegment):
    __slots__ = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.duration_ms = self.duration_seconds * 1000.

    def append(self, seg, _=None):
        """pydub.AudioSegment. Without cross-fade."""
        seg1, seg2 = AudioSegment._sync(self, seg)
        # noinspection PyProtectedMember
        return seg1._spawn(seg1._data + seg2._data)

    __add__ = append


def speed_change(audio: AudioSegment, speed_changes: float, allow_copy: bool = True, log_level: str = 'error'
                 ) -> AudioSegment:
    if speed_changes == 1.:
        return audio if allow_copy else copy(audio)
    if speed_changes <= 0.:
        raise ValueError(f"Speed cannot be negative ({speed_changes}).\n"
                         "This is usually due to errors in subtitle timecodes.")

    atempo = array('d')
    if speed_changes < .5:
        _n = speed_changes
        while _n < .5:
            atempo.append(.5)
            _n *= 2.
        atempo.append(_n)
    elif speed_changes > 100.:
        _n = speed_changes
        while _n > 100.:
            atempo.append(100.)
            _n /= 2.
        atempo.append(_n)
    else:
        atempo.append(speed_changes)

    with TemporaryDirectory() as tmp:
        inp = os.path.join(tmp, 'inp.mp3')
        audio.export(inp)
        out = os.path.join(tmp, 'out.mp3')
        FFMpegWrapper.convert('-i', inp,
                              '-af', ','.join([f'atempo={i}' for i in atempo]),
                              out, loglevel=log_level)
        return AudioSegment.from_file(out)


def fit(audio: AudioSegment,
        left_border: float, need_duration: float, right_border: float,
        align: float) -> AudioSegment:
    """Fits audio to the borders of the subtitles."""
    if (audio_duration := audio.duration_ms) > (free := (left_border + need_duration + right_border)):
        audio = speed_change(audio, audio_duration / free)
        audio_duration = audio.duration_ms
    return AudioSegment.silent(
        (free - audio_duration) / align
        if audio_duration > need_duration + right_border
        else left_border) + audio


def side_chain(sound1: AudioSegment, sound2: AudioSegment,
               min_silence_len: int = 100, silence_thresh: float = -inf,
               gain_during_overlay: int = -10
               ) -> AudioSegment:
    for start, end in tqdm(
            pydub.silence.detect_nonsilent(sound2, min_silence_len=min_silence_len, silence_thresh=silence_thresh),
            desc="Ducking"):
        sound1 = sound1.overlay(sound2[start:end], position=start, gain_during_overlay=gain_during_overlay)
    return sound1



from __future__ import annotations

__all__ = ('PrettyViewPrefix',)


class PrettyViewPrefix:
    """Class for representing a number as units"""
    __slots__ = ()

    @staticmethod
    def from_any(size: float, div: float, base: str, prefixes) -> str:
        for prefix in prefixes[:-1]:
            if size < div:
                return f'{size:,g}{prefix}{base}'
            size /= div
        return f'{size:,g}{prefixes[-1]}{base}'

    @classmethod
    def from_bytes(cls, size: float) -> str:
        return cls.from_any(size, 1000., 'B', ('', 'k', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'))

    @classmethod
    def from_seconds(cls, size: float) -> str:
        return cls.from_any(size, 60., '', 'smh')



from __future__ import annotations

import datetime
import os.path
import re

from fastdub.ffmpeg_wrapper import FFMpegWrapper

__all__ = ('LINE_REGEX',
           'Line',
           'parse', 'unparse',
           'ms_to_srt_time')

LINE_REGEX = re.compile(r'\n\n^\d+$\n', re.M)


def ms_to_srt_time(ms: int) -> str:
    s, ms = ms // 1000, ms % 1000
    m, s = s // 60, s % 60
    h, m = m // 60, m % 60
    return f'{h:0>2.0f}:{m:0>2.0f}:{s:0>2.0f},{ms:.0f}'


class Line:
    __slots__ = ('ms', 'text', '_as_repr')

    class TimeLabel:
        __slots__ = ('start', 'duration', 'end', 'time_str')

        def __init__(self, start: int, end: int):
            self.duration = end - start
            self.start = start
            self.end = end
            self.time_str = f'{start} --> {end}'

        def __str__(self):
            return (f"{ms_to_srt_time(self.start)}"
                    " --> "
                    f"{ms_to_srt_time(self.end)}")

    def __init__(self, time_labels: tuple[datetime.time], text: str):
        self.ms: Line.TimeLabel = self.TimeLabel(*
                                                 [int(
                                                     label.hour * 3600000
                                                     + label.minute * 60000
                                                     + label.second * 1000
                                                     + label.microsecond / 1000
                                                 ) for label in (time_labels[0], time_labels[-1])][:2])
        self.text: str = text
        self._as_repr = f'Line({self.ms}, {self.text!r})'

    def __repr__(self):
        return self._as_repr


def parse(text_or_file: str, skip_empty: bool = False) -> tuple[Line] | tuple:
    if os.path.isfile(text_or_file):
        fn, ext = os.path.splitext(text_or_file)
        if ext != '.srt':
            converted = f'{fn}.srt'
            FFMpegWrapper.convert('-i', text_or_file, converted)
            text_or_file = converted
        with open(text_or_file, encoding='UTF-8') as f:
            text = f.read()
    else:
        text = text_or_file
    subtitles = ()
    for i in LINE_REGEX.split(f'\n\n{text.lstrip()}')[1:]:
        times_text: list[str, str] = i.split('\n', 1)
        text = times_text[1].strip()
        if not text:
            continue
        subtitles += Line((*(datetime.time(k.hour, k.minute, k.second, k.microsecond) for k in
                             [datetime.datetime.strptime(j, '%H:%M:%S,%f') for j in
                              times_text[0].split(' --> ', 1)]),), text),
    return (*(line for line in subtitles if line.text.strip()),) if skip_empty else subtitles


def unparse(subtitles: tuple[Line]) -> str:
    return '\n\n'.join(f'{i}\n{line.ms}\n{line.text}'
                       for i, line in enumerate(sorted(subtitles, key=lambda k: k.ms.start), 1))



from pathlib import Path

import setuptools

with open('requires.txt', encoding='utf-8') as f:
    requires = f.read().strip().splitlines()
with open('CHANGELOG.md', encoding='utf-8') as f:
    changelog = f.read()
with open('README.md', encoding='utf-8') as f:
    readme = f.read()
readme += f'\n\n# CHANGELOG\n\n{changelog}'

extras_require = {}
for require in Path('extra_requires').glob('*_*.txt'):
    with open(require, encoding='UTF-8') as f:
        extras_require[require.name.split('.', 1)[0].rsplit('_', 1)[-1].casefold()] = f.read().strip().splitlines()
extras_require['all'] = sum(extras_require.values(), requires)

setuptools.setup(
    name="FastDub",
    version="3.2.0",

    description="A Python CLI package "
                "for voice over subtitles, with the ability to embed in video, audio ducking, "
                "and dynamic voice changer for a single track; "
                "auto translating; "
                "download and upload to YouTube supports",

    long_description=readme,
    long_description_content_type="text/markdown",

    author="Nikita (NIKDISSV)",
    author_email='nikdissv@proton.me',

    packages=setuptools.find_packages(),

    install_requires=requires,
    extras_require=extras_require,

    project_urls={
        'Download Voices': 'https://rhvoice.su/voices/',
        'GitHub': 'https://github.com/NIKDISSV-Forever/FastDub/',
        'YouTube': 'https://www.youtube.com/channel/UC8JV3zPSVm9EKSWD1XdkQvw/',
    },

    license='MIT',
    python_requires='>=3.8',

    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Environment :: Console',

        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Other Audience',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Telecommunications Industry',

        'License :: OSI Approved',
        'License :: OSI Approved :: MIT License',

        'Operating System :: OS Independent',

        'Natural Language :: English',
        'Natural Language :: Russian',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',

        'Topic :: Multimedia',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Multimedia :: Sound/Audio :: Analysis',
        'Topic :: Multimedia :: Sound/Audio :: Conversion',
        'Topic :: Multimedia :: Sound/Audio :: Speech',
        'Topic :: Multimedia :: Video',
        'Topic :: Multimedia :: Video :: Conversion',
        'Topic :: Software Development :: Internationalization',
        'Topic :: Utilities',

        'Typing :: Typed',
    ],

    keywords=['dubbing', 'voicing',
              'fastdub', 'JustDub', 'SpeedDub',
              'offline', 'free', 'easiest',
              'subtitles', 'videos', 'veed', 'fast'],
)



from pathlib import Path
from typing import Callable

from fastdub.youtube import pafy
from fastdub.youtube.downloader import with_api_key
from fastdub.youtube.yt_upload.upload_video import upload


class Uploader:
    def __init__(self,
                 privacy_status: str = 'private',
                 translate: bool = False, translate_service: Callable = None, ):
        self.translate = (
            (lambda line, lang: translate_service(line, to_language=lang))
            if translate else (lambda line, _: line))
        self.privacy = privacy_status

    def upload(self, base_dir: Path):
        names = {fn: (fn.parent.name.removesuffix('_result').removeprefix('_'), fn.stem) for fn in
                 base_dir.glob('_*_result/*.mp4')}
        for fp, (name, lang) in names.items():
            def translate(text: str) -> str:
                return self.translate(text, lang)

            title = f'{name} - {lang}'
            description = category = keywords = None
            ytdl = None
            try:
                ytdl = with_api_key(lambda: pafy.new(name))
            except (OSError, ValueError):
                pass
            if ytdl:
                title = ytdl.title
                description = ytdl.description
                keywords = ytdl.keywords
            command = ('--file', str(fp), '--title', translate(title),
                       '--privacy-status', self.privacy, '--noauth_local_webserver')
            if description:
                command += ('--description', translate(description),)
            if category:
                command += ('--category', category)
            if keywords:
                command += ('--keywords', ','.join(translate(kw) for kw in keywords))
            upload(command)



from fastdub.youtube import yt_upload

__all__ = ('SUPPORTED',)
try:
    from fastdub.youtube import downloader
    from fastdub.youtube import subtitles
    from fastdub.youtube.pafy.backend_youtube_dl import YtdlPafy
except ImportError as e:
    from typing import Any as YtdlPafy

    SUPPORTED = False
else:
    SUPPORTED = True
    __all__ += ('downloader', 'subtitles', 'YtdlPafy')
